package graph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;

import org.jblas.FloatMatrix;
import org.jgrapht.Graph;

import solvers.DirtyAddVertex;
import Jama.Matrix;
import dataStructures.Store;
import distanceFunction.SymmetricDistanceFunction;

public class ConflictCachingGraph<V, E> extends VertexMaskedSubgraph<V, E>
{
	class ConflictStore
	{
		/**
		 * Stores for each vertex v the vertices that are in conflict with v.
		 */
		private final HashMap<V, Set<V>> conflictEntries;

		/**
		 * Stores for each vertex pair how many neighbors they have in common.
		 */
		private final SymmetricDistanceFunction<V, Integer> noCommonNeigbors;

		public ConflictStore()
		{
			this(null);
		}

		protected ConflictStore(ConflictStore store)
		{
			SimpleMatrix commonNeighborhoodMatrix;
			final ArrayList<V> vertexSet = new ArrayList<V>(ConflictCachingGraph.this.vertexSet());

			if ( ConflictCachingGraph.this.vertexSet().size() > 800 )
			{
				final float[][] mats = new float[vertexSet.size()][];
				for (int i = 0; i < vertexSet.size(); i++)
				{
					mats[i] = new float[vertexSet.size()];
				}
				for (int i = 0; i < vertexSet.size(); i++)
				{
					for (int j = i + 1; j < vertexSet.size(); j++)
					{
						float value = 0f;
						if ( ConflictCachingGraph.this.containsEdge(vertexSet.get(i), vertexSet.get(j)) )
						{
							value = 1f;
						}
						mats[i][j] = value;
						mats[j][i] = value;
					}
				}

				class FloatMatrixWrapper implements SimpleMatrix
				{
					private final FloatMatrix mat;

					public FloatMatrixWrapper(FloatMatrix matrix)
					{
						this.mat = matrix;
					}

					@Override
					public int get(int i, int j)
					{
						return (int) this.mat.get(i, j);
					}

				}
				;

				final FloatMatrix mat2 = new FloatMatrix(mats);
				commonNeighborhoodMatrix = new FloatMatrixWrapper(mat2.mmul(mat2));
			}
			else
			{
				final double[][] mats = new double[vertexSet.size()][];
				for (int i = 0; i < vertexSet.size(); i++)
				{
					mats[i] = new double[vertexSet.size()];
				}
				for (int i = 0; i < vertexSet.size(); i++)
				{
					for (int j = i + 1; j < vertexSet.size(); j++)
					{
						double value = 0d;
						if ( ConflictCachingGraph.this.containsEdge(vertexSet.get(i), vertexSet.get(j)) )
						{
							value = 1d;
						}
						mats[i][j] = value;
						mats[j][i] = value;
					}
				}

				final Matrix mat2 = new Matrix(mats);
				class MatrixWrapper implements SimpleMatrix
				{
					private final Matrix ma;

					public MatrixWrapper(Matrix mat)
					{
						this.ma = mat;
					}

					@Override
					public int get(int i, int j)
					{
						return (int) this.ma.get(i, j);
					}

				}
				;
				commonNeighborhoodMatrix = new MatrixWrapper(mat2.times(mat2));

			}

			this.noCommonNeigbors = store == null ? new SymmetricDistanceFunction<V, Integer>(vertexSet, 0)
					: store.noCommonNeigbors;
			this.conflictEntries = new HashMap<V, Set<V>>((int) (vertexSet.size() * 1.4));

			for (final V v : vertexSet)
			{
				this.conflictEntries.put(v, new HashSet<V>());
			}

			for (int i = 0; i < vertexSet.size(); i++)
			{
				final V vi = vertexSet.get(i);
				final Set<V> conflictSetVi = this.conflictEntries.get(vi);

				for (int j = i + 1; j < vertexSet.size(); j++)
				{
					final V vj = vertexSet.get(j);
					final int noCommonNeighbors = commonNeighborhoodMatrix.get(i, j);

					this.noCommonNeigbors.set(vi, vj, noCommonNeighbors);
					if ( noCommonNeighbors <= 0 && !ConflictCachingGraph.this.containsEdge(vi, vj) )
					{
						conflictSetVi.add(vj);
						this.conflictEntries.get(vj).add(vi);
					}
				}
			}
		}

		public void addVertex(DirtyAddVertex<V> addV)
		{
			this.conflictEntries.put(addV.getVertex(), addV.getVerticesInConflict());
			for (final V con : addV.getVerticesInConflict())
			{
				this.conflictEntries.get(con).add(addV.getVertex());
			}

			for (final Store<V, V> neighDecreases : addV.getNeighborDecrease())
			{
				this.noCommonNeigbors.set(neighDecreases.getS1(), neighDecreases.getS2(),
						this.noCommonNeigbors.get(neighDecreases.getS1(), neighDecreases.getS2()) + 1);
			}

			for (final Store<V, V> newConflicts : addV.getNewConflicts())
			{
				this.conflictEntries.get(newConflicts.getS1()).remove(newConflicts.getS2());
				this.conflictEntries.get(newConflicts.getS2()).remove(newConflicts.getS1());
			}

			// this.updateCommonNeighborsAfterVertexAdd(addV.getVertex());
		}

		public void addVertex(V addV)
		{
			final Set<V> vertexSet = ConflictCachingGraph.this.vertexSet();
			this.conflictEntries.put(addV, new HashSet<V>());
			for (final V v : vertexSet)
			{
				if ( addV != v )
				{
					this.noCommonNeigbors.set(addV, v, GraphAnalysis.getCommonNeighbors(ConflictCachingGraph.this, addV, v)
							.size());

					if ( this.noCommonNeigbors.get(addV, v) <= 0 && !ConflictCachingGraph.this.containsEdge(addV, v) )
					{
						this.conflictEntries.get(addV).add(v);
						this.conflictEntries.get(v).add(addV);
					}
				}
			}
			this.updateCommonNeighborsAfterVertexAdd(addV);
		}

		private void check()
		{
			for (final V v1 : ConflictCachingGraph.this.vertexSet())
			{
				for (final V v2 : ConflictCachingGraph.this.vertexSet())
				{
					if ( v1 != v2 )
					{
						final Set<V> commonNeighbors = GraphAnalysis.getCommonNeighbors(ConflictCachingGraph.this, v1, v2);
						if ( this.noCommonNeigbors.get(v1, v2) != commonNeighbors.size() )
							throw new IllegalAccessError("noCommon " + v1 + "|" + v2 + " " + this.noCommonNeigbors.get(v1, v2)
									+ " " + commonNeighbors);

						if ( ConflictCachingGraph.this.containsEdge(v1, v2) || this.noCommonNeigbors.get(v1, v2) > 0 )
						{
							if ( this.conflictEntries.get(v1).contains(v2) )
								throw new IllegalAccessError("conflict1");
						}
						else if ( !this.conflictEntries.get(v1).contains(v2) )
						{
							final String errorMessage = "conflict: conflictstore for " + v1 + " does not contain " + v2;
							throw new IllegalAccessError(errorMessage);
						}
					}
					if ( this.conflictEntries.get(v1).contains(v1) )
						throw new IllegalAccessError("corrupt conflicts");
				}
			}
		}

		public HashMap<V, Set<V>> getConflictMapping()
		{
			return this.conflictEntries;
		}

		public ConflictSet<V> getLargestConflict()
		{
			final Comparator<Entry<V, Set<V>>> comp = new Comparator<Entry<V, Set<V>>>() {

				@Override
				public int compare(Entry<V, Set<V>> o1, Entry<V, Set<V>> o2)
				{
					return o1.getValue().size() - o2.getValue().size();
				}

			};
			final Set<Entry<V, Set<V>>> entrySet = this.conflictEntries.entrySet();
			if ( entrySet == null || entrySet.size() == 0 )
				return null;
			final Entry<V, Set<V>> e = Collections.max(entrySet, comp);

			if ( e == null || e.getValue().size() == 0 )
				return null;
			return new ConflictSet<V>(e.getKey(), e.getValue());
		}

		public DirtyAddVertex<V> removeVertex(V vRemove)
		{
			final Set<V> conflict = this.conflictEntries.remove(vRemove);
			final DirtyAddVertex<V> dirtyRemove = new DirtyAddVertex<V>(vRemove, conflict);

			for (final V conflictVertices : conflict)
			{
				this.conflictEntries.get(conflictVertices).remove(vRemove);
			}

			// decrease the number of common neighbors for each neighborhood pair by one
			final ArrayList<V> arrayList = new ArrayList<V>(ConflictCachingGraph.this.getNeighborsOf(vRemove));
			for (int i = 0; i < arrayList.size(); i++)
			{
				final V n1 = arrayList.get(i);
				for (int j = i + 1; j < arrayList.size(); j++)
				{
					final V n2 = arrayList.get(j);
					final int value = this.noCommonNeigbors.get(n1, n2) - 1;

					this.noCommonNeigbors.set(n1, n2, value);
					dirtyRemove.addCommonNeighDecrease(n1, n2);

					if ( value == 0 && !ConflictCachingGraph.this.containsEdge(n1, n2) )
					{
						this.conflictEntries.get(n1).add(n2);
						this.conflictEntries.get(n2).add(n1);
						dirtyRemove.addNewConflict(n1, n2);
					}
				}
			}
			return dirtyRemove;
		}

		private void updateCommonNeighborsAfterVertexAdd(V addV)
		{
			// increase the number of common neighbors for each neighborhood pair by one
			final ArrayList<V> neighbors = new ArrayList<V>(ConflictCachingGraph.this.getNeighborsOf(addV));
			// final ArrayList<V> neighbors = new
			// ArrayList<V>(GraphAnalysis.neigbors(ConflictCachingGraph.this, addV));
			for (int i = 0; i < neighbors.size(); i++)
			{
				for (int j = i + 1; j < neighbors.size(); j++)
				{
					final V n1 = neighbors.get(i);
					final V n2 = neighbors.get(j);
					Integer value = this.noCommonNeigbors.get(n1, n2);
					if ( value == null || value.intValue() == 0 )
					{
						value = 1;
						this.conflictEntries.get(n1).remove(n2);
						this.conflictEntries.get(n2).remove(n1);
					}
					else
					{
						value++;
					}
					this.noCommonNeigbors.set(n1, n2, value);
				}
			}
		}
	}

	public interface SimpleMatrix
	{
		public int get(int i, int j);
	}

	private final ConflictStore conflictstore;

	public ConflictCachingGraph(Graph<V, E> graph, Set<V> vertexSet)
	{
		super(graph, vertexSet);
		this.conflictstore = new ConflictStore();
	}

	@Override
	public boolean addVertex(DirtyAddVertex<V> vertex)
	{

		final boolean add = super.addVertex(vertex);
		this.conflictstore.addVertex(vertex);

		return add;
	}

	@Override
	public boolean addVertex(V v)
	{

		final boolean addVertex = super.addVertex(v);
		this.conflictstore.addVertex(v);

		return addVertex;
	}

	@Override
	public void checkConsistency()
	{
		super.checkConsistency();
		this.conflictstore.check();
	}

	public HashMap<V, Set<V>> getConflictMapping()
	{
		return this.conflictstore.getConflictMapping();
	}

	public Set<V> getConflictVertices(V v)
	{
		return this.getConflictMapping().get(v);
	}

	public ConflictSet<V> getLargeConflict()
	{
		return this.conflictstore.getLargestConflict();
	}

	public V getMinTwoNeighVertex()
	{
		final ConflictSet<V> largestConflict = this.conflictstore.getLargestConflict();
		return largestConflict != null ? largestConflict.getV() : (this.vertexSet().size() > 0 ? this.vertexSet().iterator()
				.next() : null);
	}

	public int getNoOfCommonNeigbors(V v1, V v2)
	{
		return this.conflictstore.noCommonNeigbors.get(v1, v2);
	}

	public Set<V> getTwoNeigh(V v)
	{
		if ( this.getTwoNeighSize(v) * 2 > this.vertexSet().size() )
		{
			final HashSet<V> hashSet = new HashSet<V>(this.vertexSet());
			hashSet.removeAll(this.getConflictMapping().get(v));
			return hashSet;
		}
		return GraphAnalysis.getNeighborsWithinDistanceS(v, this, 2);
	}

	public int getTwoNeighSize(V v)
	{
		final Set<V> set = this.getConflictMapping().get(v);
		return this.vertexSet().size() - (set != null ? set.size() : 0);
	}

	public DirtyAddVertex<V> removeDirtyVertex(V arg0)
	{

		final DirtyAddVertex<V> conflict = this.conflictstore.removeVertex(arg0);
		conflict.setNeighbors(super.removeVertexAndGetOldNeighbors(arg0));

		return conflict;
	}

	@Override
	public boolean removeVertex(V arg0)
	{

		this.conflictstore.removeVertex(arg0);
		final boolean removeVertex = super.removeVertex(arg0);

		return removeVertex;
	}

}
