package graph;

import graphana.graphs.defaulttypes.StdVertex;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import libraries.jgrapht.JGraphTWeightedStatusEdge;
import nl.uu.cs.treewidth.algorithm.AllStartLexBFS;
import nl.uu.cs.treewidth.algorithm.AllStartMaximumCardinalitySearch;
import nl.uu.cs.treewidth.algorithm.AllStartMaximumMinimumDegree;
import nl.uu.cs.treewidth.algorithm.AllStartMaximumMinimumDegreePlusLeastC;
import nl.uu.cs.treewidth.algorithm.AllStartMinorMinWidth;
import nl.uu.cs.treewidth.algorithm.GreedyDegree;
import nl.uu.cs.treewidth.algorithm.GreedyFillIn;
import nl.uu.cs.treewidth.algorithm.LowerBound;
import nl.uu.cs.treewidth.algorithm.MaximumCardinalitySearch;
import nl.uu.cs.treewidth.algorithm.MaximumMinimumDegree;
import nl.uu.cs.treewidth.algorithm.MaximumMinimumDegreePlusLeastC;
import nl.uu.cs.treewidth.algorithm.MaximumMinimumDegreePlusMaxD;
import nl.uu.cs.treewidth.algorithm.MaximumMinimumDegreePlusMinD;
import nl.uu.cs.treewidth.algorithm.MinDegree;
import nl.uu.cs.treewidth.algorithm.MinorMinWidth;
import nl.uu.cs.treewidth.algorithm.Ramachandramurthi;
import nl.uu.cs.treewidth.algorithm.TreewidthDP;
import nl.uu.cs.treewidth.input.GraphInput.InputData;
import nl.uu.cs.treewidth.ngraph.ListGraph;
import nl.uu.cs.treewidth.ngraph.ListVertex;
import nl.uu.cs.treewidth.ngraph.NGraph;
import nl.uu.cs.treewidth.ngraph.NVertex;

import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.SimpleGraph;

import benchmark.helpClasses.LogFileWriter;
import dataStructures.Store;

public class GraphAnalysis
{

	static class DegreeComparator<V, E> implements Comparator<V>
	{

		private final Graph<V, E> g;

		public DegreeComparator(Graph<V, E> graph)
		{
			this.g = graph;
		}

		@Override
		public int compare(V o1, V o2)
		{
			return GraphAnalysis.neigbors(this.g, o2).size() - GraphAnalysis.neigbors(this.g, o1).size();
		}
	}

	public static <V, E> Set<V> getAllNonKAnonymousVertices(VertexMaskedSubgraph<V, E> graph, int k)
	{
		TreeMap<Integer, Set<V>> degreeS = new TreeMap<Integer, Set<V>>();
		for (V v : graph.vertexSet())
		{
			int deg = graph.getDegreeOf(v);
			if ( !degreeS.containsKey(deg) )
			{
				degreeS.put(deg, new HashSet<V>());

			}
			if ( degreeS.get(deg).size() < k )
				degreeS.get(deg).add(v);
		}
		Set<V> nonKAnynomousVertices = new HashSet<V>();
		for (Integer i : degreeS.keySet())
		{
			Set<V> set = degreeS.get(i);
			if ( set.size() < k )
			{
				nonKAnynomousVertices.addAll(set);
			}
		}
		return nonKAnynomousVertices;
	}

	public static void analyseGraph(final LogFileWriter writer, final Graph graph, final boolean treewidth) throws IOException
	{
		final int n = graph.vertexSet().size();
		final int m = graph.edgeSet().size();

		writer.write("n:\t" + n);

		writer.write("\tm:\t" + m);
		writer.write("\t#FES:\t" + (m - (n - 1)));
		writer.flush();
		final int maxDegree = GraphAnalysis.maxDegree(graph);
		writer.write("\tmax deg:\t" + maxDegree);
		writer.flush();

		final float avgDegree = GraphAnalysis.getAverageDegree(graph);
		writer.write("\tavg deg:\t" + avgDegree);

		// final int vc = VertexCovers.find2ApproximationCover(graph).size();
		// // ;Math.min(VertexCovers.findGreedyCover(internal).size(),
		// // VertexCovers.find2ApproximationCover(internal)
		// writer.write("\tvc size: \t" + vc);
		// writer.flush();
		//
		// GraphanaAccess graphanaAccess = new GraphanaAccess();
		// graphanaAccess.setJGraphGraph(graph); // or JUNG, or general GraphLibrary
		// writer.write("\tdegeneracy:\t"+graphanaAccess.executeInt("degeneracy"));

		writer.write("\t h-index:\t" + GraphAnalysis.getHIndex(graph));
		writer.flush();

		if ( treewidth )
		{
			final int tw = GraphAnalysis.getTreewidthUpperBound(graph);
			System.gc();
			writer.write("\ttw UB:\t" + tw);

			final int twLB = GraphAnalysis.getTreewidthLowerBound(graph);
			writer.write("\ttw LB:\t" + twLB);
		}
	}

	public static NGraph<InputData> convertJGraphtToNGraph(final Graph internal)
	{
		final NGraph<InputData> graph = new ListGraph<InputData>();
		final NVertex<InputData> vertexPrototype = new ListVertex<InputData>();

		final HashMap<Object, NVertex<InputData>> map = new HashMap<Object, NVertex<InputData>>();

		int i = 0;
		for (final Object vertex : internal.vertexSet())
		{
			final NVertex<InputData> v = vertexPrototype.newOfSameType(new InputData(i, Integer.toString(i++)));
			graph.addVertex(v);
			map.put(vertex, v);
		}

		for (final Object edge : internal.edgeSet())
		{
			graph.addEdge(map.get(internal.getEdgeSource(edge)), map.get(internal.getEdgeTarget(edge)));
		}
		return graph;
	}

	public static SimpleGraph<StdVertex, DefaultEdge> createPGeneratedGraph(int vertexCount, float a, float b)
	{
		final SimpleGraph<StdVertex, DefaultEdge> graph = new SimpleGraph<StdVertex, DefaultEdge>(DefaultEdge.class);

		final float[] p = new float[vertexCount];

		final ArrayList<StdVertex> vertices = new ArrayList<StdVertex>();
		for (int i = 0; i < vertexCount; i++)
		{
			final StdVertex stdVertex = new StdVertex("v" + i);
			graph.addVertex(stdVertex);
			vertices.add(stdVertex);
			p[i] = (float) (Math.random() * (b - a) + a);
		}

		for (int i = 0; i < vertices.size(); i++)
		{
			for (int j = i + 1; j < vertices.size(); j++)
			{
				if ( Math.random() < (p[i] + p[j]) / 2 )
				{
					graph.addEdge(vertices.get(i), vertices.get(j));
				}
			}
		}
		return graph;
	}

	public static <V, E> int degreeOf(Graph<V, E> graph, V vertex)
	{
		if ( graph instanceof ExtendedGraph )
			return ((ExtendedGraph<V, E>) graph).getDegreeOf(vertex);
		else
			return GraphAnalysis.neigbors(graph, vertex).size();
	}

	public static <V, E> float getAverageDegree(Graph<V, E> graph)
	{
		if ( graph.vertexSet().size() > 0 )
		{
			int degree = 0;
			for (final V v : graph.vertexSet())
			{
				degree += GraphAnalysis.degreeOf(graph, v);
			}
			return (float) (degree / (double) graph.vertexSet().size());
		}
		return -1f;

	}

	public static <V, E> Set<V> getCommonNeighbors(Graph<V, E> graph, V v1, V v2)
	{
		final Set<V> n1 = GraphAnalysis.neigbors(graph, v1);
		final Set<V> n2 = GraphAnalysis.neigbors(graph, v2);
		return GraphAnalysis.getIntersection(n1, n2);

	}

	/**
	 * Import DIMACS format
	 * 
	 * @param filePath
	 * @return
	 * @throws IOException
	 */
	public static SimpleGraph<StdVertex, DefaultEdge> getDimacsGraphFromFile(String filePath) throws IOException
	{
		final SimpleGraph<StdVertex, DefaultEdge> graph = new SimpleGraph<StdVertex, DefaultEdge>(DefaultEdge.class);
		final Map<String, StdVertex> intToVertexMap = new TreeMap<String, StdVertex>();

		final File file = new File(filePath);
		if ( !file.isDirectory() )
		{
			final LineNumberReader reader = new LineNumberReader(new FileReader(file));
			String str;
			while ((str = reader.readLine()) != null)
			{
				str = str.trim();
				if ( !str.startsWith("#") )
				{
					final String[] vertices = str.split("\t");
					if ( vertices != null && vertices.length == 2 )
					{
						final String vert1 = vertices[0];
						if ( !intToVertexMap.containsKey(vert1) )
						{
							final StdVertex stdVertex = new StdVertex(vertices[0]);
							intToVertexMap.put(vert1, stdVertex);
							graph.addVertex(stdVertex);
						}
						final String vert2 = vertices[1];
						if ( !intToVertexMap.containsKey(vert2) )
						{
							final StdVertex stdVertex = new StdVertex(vertices[1]);
							intToVertexMap.put(vert2, stdVertex);
							graph.addVertex(stdVertex);
						}
						graph.addEdge(intToVertexMap.get(vert1), intToVertexMap.get(vert2));
					}
				}
			}
			reader.close();
		}
		return graph;
	}

	public static SimpleGraph<StdVertex, DefaultEdge> getEmailGraphKarlsruhe(String filePath) throws IOException
	{
		final SimpleGraph<StdVertex, DefaultEdge> graph = new SimpleGraph<StdVertex, DefaultEdge>(DefaultEdge.class);
		final Map<String, StdVertex> intToVertexMap = new TreeMap<String, StdVertex>();

		final File file = new File(filePath);
		if ( !file.isDirectory() )
		{
			final LineNumberReader reader = new LineNumberReader(new FileReader(file));
			String str;
			while ((str = reader.readLine()) != null)
			{
				str = str.trim();
				final String[] split = str.split(";");
				if ( split.length == 6 )
				{
					final String sender = split[1].trim();
					final String receiver = split[3].trim();
					if ( !sender.equalsIgnoreCase(receiver) )
					{
						if ( !intToVertexMap.containsKey(sender) )
						{
							final StdVertex stdVertex = new StdVertex("v:" + sender + ";chair:" + split[2].trim());
							intToVertexMap.put(sender, stdVertex);
							graph.addVertex(stdVertex);

						}
						if ( !intToVertexMap.containsKey(receiver) )
						{
							final StdVertex stdVertex = new StdVertex("v:" + receiver + ";chair:" + split[4].trim());
							intToVertexMap.put(receiver, stdVertex);
							graph.addVertex(stdVertex);
						}
						graph.addEdge(intToVertexMap.get(sender), intToVertexMap.get(receiver));
					}
				}

			}
			reader.close();
		}
		return graph;
	}

	public static int getExactTreewidth(final SimpleGraph<StdVertex, JGraphTWeightedStatusEdge> internal, int upperBound)
	{
		final TreewidthDP<InputData> twdp = new TreewidthDP<InputData>(upperBound);
		twdp.setInput(GraphAnalysis.convertJGraphtToNGraph(internal));
		twdp.run();
		upperBound = twdp.getTreewidth();
		return upperBound;
	}

	public static <V, E> int getHIndex(Graph<V, E> internal)
	{
		final ArrayList<V> vertices = new ArrayList<V>(internal.vertexSet());

		Collections.sort(vertices, new DegreeComparator<V, E>(internal));

		int h = 0;
		for (int i = 0; i < vertices.size(); i++)
		{
			if ( GraphAnalysis.degreeOf(internal, vertices.get(i)) >= (i + 1) )
			{
				h++;
			}
			else
			{
				break;
			}
		}
		return h;
	}

	public static <V, E> Set<V> getIntersection(Set<V> n1, Set<V> n2)
	{
		final HashSet<V> commonN = new HashSet<V>();

		if ( n1.size() > n2.size() )
		{
			final Set<V> tmp = n1;
			n1 = n2;
			n2 = tmp;
		}
		for (final V vT1 : n1)
		{
			if ( n2.contains(vT1) )
			{
				commonN.add(vT1);
			}
		}
		return commonN;
	}

	public static SimpleGraph<StdVertex, DefaultEdge> getJGraphTGraphFromMetisFile(final String filePath) throws IOException
	{
		final SimpleGraph<StdVertex, DefaultEdge> graph = new SimpleGraph<StdVertex, DefaultEdge>(DefaultEdge.class);

		final File file = new File(filePath);
		if ( !file.isDirectory() )
		{
			final LineNumberReader reader = new LineNumberReader(new FileReader(file));
			String str;
			int i = 0;
			StdVertex[] vertices = null;
			while ((str = reader.readLine()) != null)
			{
				str = str.trim();
				if ( i++ > 0 )
				{
					final String[] numbers = str.split(" ");
					for (final String nuString : numbers)
					{
						try
						{
							final int neigh = Integer.valueOf(nuString);
							if ( (i - 2) != (neigh - 1) )
							{
								graph.addEdge(vertices[i - 2], vertices[neigh - 1]);
							}
						} catch (final NumberFormatException e)
						{

						}
					}
				}
				else
				{
					final String[] numbers = str.split(" ");
					// ini vertices
					final Integer numberOfVertices = Integer.valueOf(numbers[0]);
					vertices = new StdVertex[numberOfVertices];
					for (int counter = 1; counter <= numberOfVertices; counter++)
					{
						final StdVertex v = new StdVertex("v" + counter);
						graph.addVertex(v);
						vertices[counter - 1] = v;
					}
				}
			}
			reader.close();
		}
		else
			throw new IllegalArgumentException("file name is a directory");

		return graph;
	}

	public static <V, E> Store<V, Set<V>> getMaxDegreeTwoClub(final Graph<V, E> graph)
	{
		final Store<V, Set<V>> twoClub = new Store<V, Set<V>>();
		for (final V vertex : graph.vertexSet())
		{
			if ( twoClub.getS2() == null || GraphAnalysis.degreeOf(graph, vertex) >= twoClub.getS2().size() )
			{
				twoClub.setS2(new HashSet<V>(GraphAnalysis.neigbors(graph, vertex)));
				twoClub.getS2().add(vertex);
				twoClub.setS1(vertex);
			}
		}
		return twoClub;
	}

	public static <V, E> V getMaxDegreeVertex(Graph<V, E> graph)
	{
		int maxDegree = -1;
		V maxDegreeVertex = null;
		{// compute maxdegree
			for (final V vertex : graph.vertexSet())
			{
				final int degreeOf = GraphAnalysis.degreeOf(graph, vertex);
				if ( degreeOf > maxDegree )
				{
					maxDegree = degreeOf;
					maxDegreeVertex = vertex;
				}
			}
		}
		return maxDegreeVertex;

	}

	public static <V, E> V getMinDegreeVertex(Graph<V, E> graph)
	{

		// if ( graph instanceof ConflictCachingGraph )
		// return ((ConflictCachingGraph<V, E>) graph).getMinDegreeVertex();

		V minVertex = null;
		for (final V v : graph.vertexSet())
		{
			if ( minVertex == null || GraphAnalysis.neigbors(graph, v).size() < GraphAnalysis.neigbors(graph, minVertex).size() )
			{
				minVertex = v;
			}
		}
		return minVertex;
	}

	public static <V, E> HashSet<V> getNeighborsWithinDistanceS(final V startVertex, final Graph<V, E> graph, final int s)
	{
		final HashSet<V> vSet = new HashSet<V>();
		vSet.add(startVertex);

		HashSet<V> dist_i_minus1_vertices = new HashSet<V>();
		dist_i_minus1_vertices.add(startVertex);

		for (int i = 1; i <= s; i++)
		{
			final HashSet<V> newVertices = new HashSet<V>();
			for (final V v : dist_i_minus1_vertices)
			{
				newVertices.addAll(GraphAnalysis.neigbors(graph, v));
			}
			vSet.addAll(newVertices);
			dist_i_minus1_vertices = newVertices;
		}
		return vSet;
	}

	// DualSearchTreeSolver dualST = new DualSearchTreeSolver(new sClub_Instance(graph, 2,
	// maxDegree + 2));
	// Set vertexSet = dualST.solve();
	// writer.writeln(vertexSet.size());
	public static int getTreewidthLowerBound(final Graph<StdVertex, JGraphTWeightedStatusEdge> internal)
	{
		int twLB = -1;
		final NGraph<InputData> convertJGraphtToNGraph = GraphAnalysis.convertJGraphtToNGraph(internal);
		for (int heur = 1; heur <= 12; heur++)
		{
			LowerBound<InputData> twAlgo;

			if ( heur == 1 )
			{
				twAlgo = new MaximumCardinalitySearch<InputData>();
			}
			else if ( heur == 2 )
			{
				twAlgo = new Ramachandramurthi<InputData>();
			}
			else if ( heur == 3 )
			{
				twAlgo = new AllStartMaximumCardinalitySearch<InputData>();
			}
			else if ( heur == 4 )
			{
				twAlgo = new MaximumMinimumDegree<InputData>();
			}
			else if ( heur == 5 )
			{
				twAlgo = new MaximumMinimumDegreePlusLeastC<InputData>();
			}
			else if ( heur == 6 )
			{
				twAlgo = new MaximumMinimumDegreePlusMaxD<InputData>();
			}
			else if ( heur == 7 )
			{
				twAlgo = new MaximumMinimumDegreePlusMinD<InputData>();
			}
			else if ( heur == 8 )
			{
				twAlgo = new AllStartMaximumMinimumDegree<InputData>();
			}
			else if ( heur == 9 )
			{
				twAlgo = new AllStartMaximumMinimumDegreePlusLeastC<InputData>();
			}
			else if ( heur == 10 )
			{
				twAlgo = new AllStartMinorMinWidth<InputData>();
			}
			else if ( heur == 11 )
			{
				twAlgo = new MinorMinWidth<InputData>();
			}
			else if ( heur == 12 )
			{
				twAlgo = new MinDegree<InputData>();
			}
			else
				throw new RuntimeException("Unknown algorithm.");

			twAlgo.setInput(convertJGraphtToNGraph);
			twAlgo.run();
			twLB = Math.max(twLB, twAlgo.getLowerBound());
		}
		return twLB;
	}

	public static int getTreewidthUpperBound(final Graph<StdVertex, JGraphTWeightedStatusEdge> internal)
	{
		final AllStartLexBFS<InputData> allStartLexBFS = new AllStartLexBFS<InputData>();
		final NGraph<InputData> nGraph = GraphAnalysis.convertJGraphtToNGraph(internal);
		allStartLexBFS.setInput(nGraph);
		allStartLexBFS.run();
		int tw = allStartLexBFS.getUpperBound();

		final GreedyFillIn<InputData> greedy = new GreedyFillIn<InputData>();
		greedy.setInput(nGraph);
		greedy.run();
		tw = Math.min(tw, greedy.getUpperBound());

		final GreedyDegree<InputData> greedy2 = new GreedyDegree<InputData>();
		greedy2.setInput(nGraph);
		greedy2.run();
		tw = Math.min(tw, greedy2.getUpperBound());
		return tw;
	}

	/**
	 * Finds an arbitrary conflict set.
	 * 
	 * @param graph
	 * @return
	 */
	public static <V, E> ConflictSet<V> getVertexPairToFarApart(Graph<V, E> graph)
	{
		for (final V v1 : graph.vertexSet())
		{
			ConflictSet<V> conflictSetV1 = null;
			for (final V v2 : graph.vertexSet())
			{
				if ( v1 != v2 )
				{
					final Set<V> n1 = GraphAnalysis.neigbors(graph, v1);
					n1.add(v1);
					final Set<V> n2 = GraphAnalysis.neigbors(graph, v2);
					n2.add(v2);
					if ( !GraphAnalysis.haveNonEmptyIntersection(n1, n2) )
					{
						if ( conflictSetV1 == null )
						{
							conflictSetV1 = new ConflictSet<V>(v1, null);
						}
						conflictSetV1.add(v2);
					}
				}
			}
			if ( conflictSetV1 != null )
				return conflictSetV1;
		}
		return null;
	}

	public static <V> boolean haveNonEmptyIntersection(Collection<V> neighbors1, Collection<V> neighbors2)
	{
		if ( neighbors2.size() < neighbors1.size() )
		{
			final Collection<V> t = neighbors1;
			neighbors1 = neighbors2;
			neighbors2 = t;
		}
		for (final V v : neighbors1)
		{
			if ( neighbors2.contains(v) )
				return true;
		}
		return false;
	}

	/**
	 * Checks whether the given graph is a 2-club. TODO improve performance
	 * 
	 * @param graph
	 * @return
	 */
	public static <V, E> boolean is2Club(Graph<V, E> graph)
	{
		for (final V v1 : graph.vertexSet())
		{
			for (final V v2 : graph.vertexSet())
			{
				if ( v1 != v2 )
				{
					final Set<V> n1 = new HashSet<V>(GraphAnalysis.neigbors(graph, v1));
					n1.add(v1);
					final Set<V> n2 = new HashSet<V>(GraphAnalysis.neigbors(graph, v2));
					n2.add(v2);
					if ( !GraphAnalysis.haveNonEmptyIntersection(n1, n2) )
						return false;

				}
			}
		}
		return true;
	}

	public static <V, E> int maxDegree(final Graph<V, E> internal)
	{
		int maxDegree = -1;
		{// compute maxdegree
			for (final V vertex : internal.vertexSet())
			{
				maxDegree = Math.max(GraphAnalysis.degreeOf(internal, vertex), maxDegree);

			}
		}
		return maxDegree;
	}

	public static <V, E> Set<V> neigbors(Graph<V, E> graphVertex, V v2)
	{
		if ( graphVertex instanceof VertexMaskedSubgraph )
			return ((VertexMaskedSubgraph<V, E>) graphVertex).getNeighborsOf(v2);

		final Set<V> neighbors = new HashSet<V>();
		for (final E edge : graphVertex.edgesOf(v2))
		{
			final V v = graphVertex.getEdgeSource(edge);
			if ( v != v2 )
			{
				neighbors.add(v);
			}
			else
			{
				neighbors.add(graphVertex.getEdgeTarget(edge));
			}
		}
		return neighbors;

	}

	public static SimpleGraph<StdVertex, DefaultEdge> readEdgeGraph(String filePath) throws IOException
	{
		final SimpleGraph<StdVertex, DefaultEdge> graph = new SimpleGraph<StdVertex, DefaultEdge>(DefaultEdge.class);

		final File file = new File(filePath);
		if ( !file.isDirectory() )
		{
			final LineNumberReader reader = new LineNumberReader(new FileReader(file));
			String str;
			final TreeMap<String, StdVertex> vertices = new TreeMap<String, StdVertex>();
			while ((str = reader.readLine()) != null)
			{
				str = str.trim();
				final String[] numbers = str.split(" ");
				if ( numbers != null && numbers.length >= 2 )
				{
					numbers[0] = numbers[0].trim();
					numbers[1] = numbers[1].trim();

					if ( !vertices.containsKey(numbers[0]) )
					{
						final StdVertex stdVertex = new StdVertex("v" + numbers[0]);
						vertices.put(numbers[0], stdVertex);
						graph.addVertex(stdVertex);

					}
					if ( !vertices.containsKey(numbers[1]) )
					{
						final StdVertex stdVertex = new StdVertex("v" + numbers[1]);
						vertices.put(numbers[1], stdVertex);
						graph.addVertex(stdVertex);
					}
					graph.addEdge(vertices.get(numbers[0]), vertices.get(numbers[1]));
				}

			}
			reader.close();
		}

		return graph;
	}

	public static <V, E> void writeGraphToDimacs(Graph<V, E> graph, String filePath) throws IOException
	{
		final File file = new File(filePath);
		final FileWriter writer = new FileWriter(file);

		for (final E edge : graph.edgeSet())
		{
			writer.write(graph.getEdgeSource(edge) + "\t" + graph.getEdgeTarget(edge) + "\n");
		}
		writer.close();

	}

}
