package graph;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.jgrapht.EdgeFactory;
import org.jgrapht.Graph;

import solvers.DirtyAddVertex;

/**
 * Provides a modifiable view on a given graph. By addVertices/removeVertices one can unmask/mask
 * vertices contained in the underlying graph. The corresponding edges will be masked/unmasked as
 * well. The underlying graph cannot be modified.
 * 
 * @author sepphar
 * 
 * @param <V>
 * @param <E>
 */
public class VertexMaskedSubgraph<V, E> implements ExtendedGraph<V, E>
{
	class EdgeSet
	{
		private final Comparator<? super V> comparator = new Comparator<V>() {

			@Override
			public int compare(V arg0, V arg1)
			{
				if ( arg1 == null )
					return -1;
				if ( arg0 == null )
					return 1;

				return EdgeSet.this.getDegreeOf(arg0) - EdgeSet.this.getDegreeOf(arg1);
			}

		};

		public final HashMap<V, Set<V>> currentNeighbors;

		public final HashMap<V, Set<V>> neighborsInUnmaskedGraph;

		// private final BinaryHeap<V> degreeHeap;

		public EdgeSet()
		{
			this.neighborsInUnmaskedGraph = new HashMap<V, Set<V>>();
			this.currentNeighbors = new HashMap<V, Set<V>>();

			// add all vertices
			for (final V v : VertexMaskedSubgraph.this.graph.vertexSet())
			{
				this.neighborsInUnmaskedGraph.put(v, new HashSet<V>());
				this.currentNeighbors.put(v, new HashSet<V>());
			}
			// add all vertices

			if ( VertexMaskedSubgraph.this.graph instanceof VertexMaskedSubgraph )
			{
				final VertexMaskedSubgraph<V, E> gTmp = (VertexMaskedSubgraph<V, E>) VertexMaskedSubgraph.this.graph;
				for (final V v : VertexMaskedSubgraph.this.unmaskedVertices)
				{
					this.neighborsInUnmaskedGraph.get(v).addAll(gTmp.getNeighborsOf(v));
					final Set<V> currentNeighbors = this.currentNeighbors.get(v);
					for (final V n : gTmp.getNeighborsOf(v))
					{
						if ( VertexMaskedSubgraph.this.unmaskedVertices.contains(n) )
						{
							currentNeighbors.add(n);
						}
					}
				}
			}
			else
			{
				for (final E edge : VertexMaskedSubgraph.this.graph.edgeSet())
				{
					final V source = VertexMaskedSubgraph.this.graph.getEdgeSource(edge);
					final V target = VertexMaskedSubgraph.this.graph.getEdgeTarget(edge);

					this.neighborsInUnmaskedGraph.get(source).add(target);
					this.neighborsInUnmaskedGraph.get(target).add(source);

					if ( VertexMaskedSubgraph.this.containsVertex(source) && VertexMaskedSubgraph.this.containsVertex(target) )
					{
						this.currentNeighbors.get(source).add(target);
						this.currentNeighbors.get(target).add(source);
					}
				}
			}
			// this.degreeHeap = new BinaryHeap<V>(this.comparator,
			// VertexMaskedSubgraph.this.vertexSet());
		}

		public void addVertex(DirtyAddVertex<V> v)
		{
			this.addVertex(v.getS1(), v.getNeighors());
		}

		public void addVertex(V v)
		{
			final HashSet<V> neighbors = new HashSet<V>();
			for (final V neighbor : this.neighborsInUnmaskedGraph.get(v))
			{
				if ( VertexMaskedSubgraph.this.containsVertex(neighbor) )
				{
					neighbors.add(neighbor);
				}
			}
			this.addVertex(v, neighbors);
		}

		private void addVertex(V v, Set<V> neighbors)
		{
			this.currentNeighbors.put(v, neighbors);
			for (final V n : neighbors)
			{
				this.currentNeighbors.get(n).add(v);
				// this.degreeHeap.valueChanged(n);
			}
			// this.degreeHeap.add(v);

		}

		private int getDegreeOf(V arg0)
		{
			return this.currentNeighbors.get(arg0).size();
		}

		// public V getMinDegreeVertex()
		// {
		// return this.degreeHeap.peek();
		// }

		public void removeAllVertices(Collection<? extends V> arg0)
		{
			for (final V v : arg0)
			{
				this.removeVertex(v);
			}
		}

		public Set<V> removeVertex(V v)
		{
			// this.degreeHeap.remove(v);

			for (final V n : this.currentNeighbors.get(v))
			{
				this.currentNeighbors.get(n).remove(v);
				// this.degreeHeap.valueChanged(n);
			}

			return this.currentNeighbors.remove(v);
		}
	}

	private static final String EDGE_IS_NOT_CONTAINED = "edge is not contained in to graph";

	private static final String GRAPH_IS_SIMPLE = "graph is simple";

	private static final String UNMODIFIABLE = "the underlying graph is unmodifiable";

	private static final String UNWEIGHTED_GRAPH = "graph is unweighted";

	private static final String VERTEX_IS_NOT_CONTAINED_IN_TO_GRAPH = "vertex is not contained in to graph ";

	private final EdgeSet edgeSet;

	private final Graph<V, E> graph;

	private final HashSet<V> unmaskedVertices;

	public VertexMaskedSubgraph(Graph<V, E> graph, Set<V> vertexSet)
	{
		this.graph = graph;
		this.unmaskedVertices = new HashSet<V>(vertexSet);
		this.edgeSet = new EdgeSet();
	}

	@Override
	public E addEdge(V arg0, V arg1)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);

	}

	@Override
	public boolean addEdge(V arg0, V arg1, E arg2)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);
	}

	public boolean addVertex(DirtyAddVertex<V> v)
	{
		this.edgeSet.addVertex(v);
		final boolean add = this.unmaskedVertices.add(v.getS1());
		return add;
	}

	@Override
	public boolean addVertex(V v)
	{
		this.edgeSet.addVertex(v);
		final boolean add = this.unmaskedVertices.add(v);
		return add;
	}

	public void checkConsistency()
	{
		int min = Integer.MAX_VALUE;
		for (final V v : this.unmaskedVertices)
		{
			min = Math.min(min, this.getDegreeOf(v));
			if ( !this.graph.containsVertex(v) )
				throw new IllegalAccessError();
			for (final V n : this.getNeighborsOf(v))
			{
				if ( !this.graph.containsEdge(v, n) )
					throw new IllegalAccessError(v + "|" + n);
				if ( !this.containsVertex(n) )
					throw new IllegalAccessError(v + "|" + n);
			}
			for (final V n : GraphAnalysis.getNeighborsWithinDistanceS(v, this.graph, 1))
			{
				if ( this.containsVertex(n) && v != n && !this.getNeighborsOf(v).contains(n) )
					throw new IllegalAccessError(v + "|" + n);
			}
		}
		// if ( min != this.getDegreeOf(this.getMinDegreeVertex()) )
		// throw new IllegalAccessError("illegal min degree  vertex");

	}

	@Override
	public boolean containsEdge(E edge)
	{
		return this.containsEdge(this.graph.getEdgeSource(edge), this.graph.getEdgeTarget(edge));
	}

	@Override
	public boolean containsEdge(V v1, V v2)
	{
		return this.containsVertex(v1) && this.containsVertex(v2) && this.graph.containsEdge(v1, v2);
	}

	@Override
	public boolean containsVertex(V v)
	{
		return this.unmaskedVertices.contains(v);
	}

	@Override
	public Set<E> edgeSet()
	{
		final Set<V> vertexSet = new HashSet<V>(this.vertexSet());
		final Set<E> edgeSet = new HashSet<E>();
		for (final V v : this.vertexSet())
		{
			for (final V n : this.getNeighborsOf(v))
			{
				if ( !vertexSet.contains(n) )
				{
					final E edge = this.getEdge(v, n);
					if ( edge == null )
						throw new IllegalAccessError();
					edgeSet.add(edge);
				}
			}
			vertexSet.remove(v);
		}
		return edgeSet;
	}

	@Override
	public Set<E> edgesOf(V arg0)
	{
		throw new UnsupportedOperationException();

		// if ( this.containsVertex(arg0) )
		// {
		// final Set<E> edges = new HashSet<E>();
		// for (final V nei : this.edgeSet.neighborsInUnmaskedGraph.get(arg0))
		// {
		// if ( this.containsVertex(nei) )
		// {
		// edges.add(this.graph.getEdge(arg0, nei));
		// }
		// }
		// return edges;
		// }
		// throw new
		// IllegalArgumentException(VertexMaskedSubgraph.VERTEX_IS_NOT_CONTAINED_IN_TO_GRAPH +
		// arg0);
	}

	@Override
	public Set<E> getAllEdges(V arg0, V arg1)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.GRAPH_IS_SIMPLE);
	}

	@Override
	public int getDegreeOf(V v)
	{
		return this.edgeSet.getDegreeOf(v);
	}

	@Override
	public E getEdge(V arg0, V arg1)
	{
		if ( this.containsVertex(arg0) && this.containsVertex(arg1) )
			return this.graph.getEdge(arg0, arg1);
		throw new IllegalArgumentException(VertexMaskedSubgraph.VERTEX_IS_NOT_CONTAINED_IN_TO_GRAPH);
	}

	@Override
	public EdgeFactory<V, E> getEdgeFactory()
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.GRAPH_IS_SIMPLE);
	}

	@Override
	public V getEdgeSource(E arg0)
	{
		final V source = this.graph.getEdgeSource(arg0);
		if ( this.containsVertex(source) )
			return source;
		throw new IllegalArgumentException(VertexMaskedSubgraph.EDGE_IS_NOT_CONTAINED);
	}

	@Override
	public V getEdgeTarget(E arg0)
	{
		final V target = this.graph.getEdgeTarget(arg0);
		if ( this.containsVertex(target) )
			return target;
		throw new IllegalArgumentException(VertexMaskedSubgraph.EDGE_IS_NOT_CONTAINED);
	}

	@Override
	public double getEdgeWeight(E arg0)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNWEIGHTED_GRAPH);
	}

	// @Override
	// public V getMinDegreeVertex()
	// {
	// return this.edgeSet.getMinDegreeVertex();
	// }

	public Set<V> getNeighborsOf(V vertex)
	{
		return Collections.unmodifiableSet(this.edgeSet.currentNeighbors.get(vertex));
	}

	@Override
	public boolean removeAllEdges(Collection<? extends E> arg0)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);
	}

	@Override
	public Set<E> removeAllEdges(V arg0, V arg1)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);
	}

	@Override
	public boolean removeAllVertices(Collection<? extends V> vertices)
	{
		this.edgeSet.removeAllVertices(vertices);
		final boolean removeAll = this.unmaskedVertices.removeAll(vertices);
		return removeAll;
	}

	@Override
	public boolean removeEdge(E arg0)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);
	}

	@Override
	public E removeEdge(V arg0, V arg1)
	{
		throw new UnsupportedOperationException(VertexMaskedSubgraph.UNMODIFIABLE);
	}

	@Override
	public boolean removeVertex(V v)
	{
		this.removeVertexAndGetOldNeighbors(v);
		return true;
	}

	public Set<V> removeVertexAndGetOldNeighbors(V v)
	{
		final Set<V> neighbors = this.edgeSet.removeVertex(v);
		this.unmaskedVertices.remove(v);
		return neighbors;
	}

	@Override
	public Set<V> vertexSet()
	{
		return this.unmaskedVertices;
	}
}
