package solvers;

import graph.ConflictCachingGraph;
import graph.ConflictSet;
import graph.GraphAnalysis;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;

import org.jgrapht.Graph;

import solvers.dataReduction.AdvancedRuleProvider;
import solvers.dataReduction.ChangeLogger;
import solvers.dataReduction.CommonNeighborsRequiredVertices;
import solvers.dataReduction.FastHighDegreeRule;
import solvers.dataReduction.HighDegreeRule;
import solvers.dataReduction.RemoveDegreeOneVertices;
import solvers.dataReduction.RuleInterface;
import solvers.dataReduction.SimpleRuleProvider;
import benchmark.helpClasses.LogFileWriter;
import benchmark.helpClasses.StopWatch;

//import solvers.dataReduction.RemoveDegreeOneVertices;

public class DualSearchTreeSolver<V, E> extends Solver<V, E>
{
	public static <V, E> boolean applyDataReductionRules(AdvancedRuleProvider<V, E> logger) throws UnsolvableException
	{
		final Collection<RuleInterface<V, E>> rules = DualSearchTreeSolver.getAdvancedRules();
		return DualSearchTreeSolver.applyDataReductionRules(rules, logger);
	}

	public static <V, E> boolean applyDataReductionRules(Collection<RuleInterface<V, E>> rules, SimpleRuleProvider<V, E> logger)
			throws UnsolvableException
	{
		int somethingModified = 2;
		boolean totalRemovedSomething = false;
		while (somethingModified > 0)
		{
			somethingModified = 0;
			for (final RuleInterface<V, E> rule : rules)
			{
				somethingModified += rule.reduce(logger) ? 1 : 0;

			}
			totalRemovedSomething = totalRemovedSomething || (somethingModified > 0);
		}
		return totalRemovedSomething;
	}

	public static <V, E> boolean applyDataReductionRules(SimpleRuleProvider<V, E> logger) throws UnsolvableException
	{
		final Collection<RuleInterface<V, E>> rules = DualSearchTreeSolver.getSimpleRules();
		return DualSearchTreeSolver.applyDataReductionRules(rules, logger);
	}

	public static <V, E> Collection<RuleInterface<V, E>> getAdvancedRules()
	{
		final Collection<RuleInterface<V, E>> rules = new Vector<RuleInterface<V, E>>();
		// rules.add(new RemoveDegreeOneVertices<V, E>());
		rules.add(new CommonNeighborsRequiredVertices<V, E>());
		rules.add(new HighDegreeRule<V, E>());
		// rules.add(new RemoveConflictsWithRequiredVertices<V, E>());

		// rules.add(new FastHighDegreeRule<V, E>());
		// rules.add(new HighDegreeRule<V, E>());

		return rules;
	}

	public static <V, E> Collection<RuleInterface<V, E>> getSimpleRules()
	{
		final Collection<RuleInterface<V, E>> rules = new Vector<RuleInterface<V, E>>();
		rules.add(new RemoveDegreeOneVertices<V, E>());
		rules.add(new FastHighDegreeRule<V, E>());

		return rules;
	}

	protected ConflictCachingGraph<V, E> branchingGraph;

	private int leaves = 0;

	private int level = 0;

	/**
	 * Best solution found, so far.
	 */
	Set<V> optSol;

	private final boolean solveToOptimal;

	public DualSearchTreeSolver(final Graph<V, E> graph, final int k)
	{
		this(graph, k, null);
	}

	public DualSearchTreeSolver(final Graph<V, E> graph, final int k, Collection<V> requiredVertex)
	{
		this(graph, k, requiredVertex, false);
	}

	public DualSearchTreeSolver(final Graph<V, E> graph, final int k, Collection<V> requiredVertex, boolean solveToOptimal)
	{
		super(graph, k, requiredVertex);
		this.solveToOptimal = solveToOptimal;
	}

	private boolean branch()
	{
		this.level++;

		if ( Solver.verbose && Solver.verboseLevel > 1 )
		{
			final int printEveryIthLevel = 5;
			if ( this.level % printEveryIthLevel == 0 || this.level <= 5 )
			{
				for (int i = 1; i <= (this.level / printEveryIthLevel) + 5; i++)
				{
					System.out.print("\t");
				}
				System.out.println("lev:" + this.level + ", noVertices:" + this.branchingGraph.vertexSet().size()
						+ ",noReqVertices:" + this.getRequiredVertices().size() + ",twoClubMinSize:"
						+ this.getTwoClubMinSizeParam());
				for (int i = 1; i <= (this.level / printEveryIthLevel) + 6; i++)
				{
					System.out.print("\t");
				}
				System.out.println(this.branchingGraph.vertexSet());
			}
		}

		final ChangeLogger<V, E> logger = new ChangeLogger<V, E>(this);
		try
		{
			DualSearchTreeSolver.applyDataReductionRules(logger);
		} catch (final UnsolvableException e)
		{
			this.cleanBranchBeforeReturn(logger, true);
			return false;
		}

		// remaining instance to small?
		if ( this.branchingGraph.vertexSet().size() < this.getTwoClubMinSizeParam() )
		{
			this.cleanBranchBeforeReturn(logger, true);
			return false;
		}

		// are there some conflicts?
		final ConflictSet<V> conflict = this.branchingGraph.getLargeConflict();
		if ( conflict == null )
		{
			this.optSol = new HashSet<V>(this.branchingGraph.vertexSet());
			if ( !this.solveToOptimal )
				return true;
			else
			{
				this.setTwoClubMinSizeParam(this.optSol.size() + 1);
				this.cleanBranchBeforeReturn(logger, true);
				return false;
			}
		}

		if ( this.branchingGraph.vertexSet().size() == this.getTwoClubMinSizeParam() )
		{
			this.cleanBranchBeforeReturn(logger, true);
			return false;
		}

		{// first branch
			if ( !this.getRequiredVertices().contains(conflict.getV()) )
			{
				final DirtyAddVertex<V> d = this.branchingGraph.removeDirtyVertex(conflict.getV());

				if ( this.branch() )
					return true;

				this.branchingGraph.addVertex(d);
			}
		}

		{// second branch
			if ( this.branchingGraph.vertexSet().size() - conflict.size() >= this.getTwoClubMinSizeParam() )
			{
				this.getRequiredVertices().add(conflict.getV());// mark vertex as required

				// delete conflicting vertices
				final ArrayList<DirtyAddVertex<V>> con = new ArrayList<DirtyAddVertex<V>>(conflict.size());
				for (final V v : conflict)
				{
					con.add(this.branchingGraph.removeDirtyVertex(v));
				}

				if ( this.reqVerticesStillContained() && this.branch() )
					return true;

				// reinsert conflicting vertices in reverse order
				for (int i = con.size() - 1; i >= 0; i--)
				{
					this.branchingGraph.addVertex(con.get(i));
				}

				this.getRequiredVertices().remove(conflict.getV());
			}
		}

		this.cleanBranchBeforeReturn(logger, false);
		return false;
	}

	private void cleanBranchBeforeReturn(ChangeLogger<V, E> ruleRemovedVertices, boolean isLeave)
	{
		this.level--;
		if ( isLeave )
		{
			this.leaves++;
		}
		if ( ruleRemovedVertices != null )
		{
			ruleRemovedVertices.reverseAllChanges();
		}
	}

	public ConflictCachingGraph<V, E> getBranchingGraph()
	{
		return this.branchingGraph;
	}

	public int getNoLeaves()
	{
		return this.leaves;
	}

	private boolean reqVerticesStillContained()
	{
		for (final V req : this.getRequiredVertices())
		{
			if ( !this.branchingGraph.containsVertex(req) )
				return false;
		}
		return true;
	}

	@Override
	public Set<V> solve()
	{
		final StopWatch watch = new StopWatch();
		watch.start();

		if ( Solver.verbose )
		{
			try
			{
				System.out.print("ST:\t");
				GraphAnalysis.analyseGraph(new LogFileWriter(), this.getGraph(), false);
				System.out.print("\tk:\t" + this.getTwoClubMinSizeParam());
				System.out.print("\treq:" + Arrays.toString(this.getRequiredVertices().toArray()));
				System.out.flush();
			} catch (final IOException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		if ( this.getGraph().vertexSet().size() < this.getTwoClubMinSizeParam() )
			return null;

		// consider max degree 2-club
		final Set<V> maxDegree2Club = GraphAnalysis.getMaxDegreeTwoClub(this.getGraph()).getS2();
		if ( maxDegree2Club.size() >= this.getTwoClubMinSizeParam() )
		{
			if ( !this.solveToOptimal )
				return maxDegree2Club;
			else
			{
				this.optSol = maxDegree2Club;
				this.setTwoClubMinSizeParam(this.optSol.size() + 1);
			}
		}

		this.branchingGraph = new ConflictCachingGraph<V, E>(this.getGraph(), this.getGraph().vertexSet());

		try
		{
			DualSearchTreeSolver.applyDataReductionRules(new ChangeLogger<V, E>(this));
		} catch (final UnsolvableException e)
		{
			return this.optSol;
		}

		if ( this.branchingGraph.vertexSet().size() < this.getTwoClubMinSizeParam() )
			return this.optSol;

		this.branch();
		if ( Solver.verbose )
		{
			watch.stopp_Print("\t branching:\t", System.out);
			System.out.print("\t #leaves:\t" + this.leaves);
		}
		return this.optSol;
	}
}
