package solvers;

import graph.GraphAnalysis;
import gurobi.GRB;
import gurobi.GRB.DoubleAttr;
import gurobi.GRB.DoubleParam;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.jgrapht.Graph;

public class ILPSolver<V, E> extends Solver<V, E>
{
	static GRBEnv env;

	static
	{
		try
		{
			ILPSolver.env = new GRBEnv(null);
			ILPSolver.env.set(GRB.IntParam.OutputFlag, 0);
			ILPSolver.env.set(DoubleParam.TimeLimit, 300);
		} catch (final GRBException e)
		{
			e.printStackTrace();
		}
	}

	public ILPSolver(Graph<V, E> graph, int k, Collection<V> reqVertices)
	{
		super(graph, k, reqVertices);
	}

	@Override
	public Set<V> solve() throws UnsolvableException
	{
		try
		{
			final GRBModel model = new GRBModel(ILPSolver.env);

			final HashMap<V, GRBVar> verticesToVariable = new HashMap<V, GRBVar>();
			// Create variables
			for (final V v : this.getGraph().vertexSet())
			{
				verticesToVariable.put(v,
						model.addVar(this.getRequiredVertices().contains(v) ? 1.0 : 0.0, 1.0, 0.0, GRB.BINARY, v.toString()));
			}

			// Integrate new variables
			model.update();

			// Set objective

			GRBLinExpr expr = new GRBLinExpr();
			for (final GRBVar var : verticesToVariable.values())
			{
				expr.addTerm(1.0, var);
			}
			model.setObjective(expr, GRB.MAXIMIZE);

			// add constraint

			final ArrayList<V> vertexList = new ArrayList<V>(this.getGraph().vertexSet());
			for (int i = 0; i < vertexList.size(); i++)
			{
				for (int j = i + 1; j < vertexList.size(); j++)
				{
					if ( !this.getGraph().containsEdge(vertexList.get(i), vertexList.get(j)) )
					{
						expr = new GRBLinExpr();

						expr.addTerm(1.0, verticesToVariable.get(vertexList.get(i)));
						expr.addTerm(1.0, verticesToVariable.get(vertexList.get(j)));

						for (final V neigh : GraphAnalysis.getCommonNeighbors(this.getGraph(), vertexList.get(i),
								vertexList.get(j)))
						{
							expr.addTerm(-1.0, verticesToVariable.get(neigh));
						}
						model.addConstr(expr, GRB.LESS_EQUAL, 1.0, null);
					}
				}
			}
			// Optimize model

			model.optimize();
			final int status = model.get(GRB.IntAttr.Status);
			if ( status == GRB.OPTIMAL )
			{
				final Set<V> club = new HashSet<V>();
				for (final V v : verticesToVariable.keySet())
				{
					if ( verticesToVariable.get(v).get(DoubleAttr.X) > 0.5d )
					{
						club.add(v);
					}
				}
				if ( club.size() >= this.getTwoClubMinSizeParam() )
					return club;
				return null;
			}
		} catch (final GRBException e)
		{
			e.printStackTrace();
		}
		throw new UnsolvableException("time out");
	}
}
