package solvers;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.jgrapht.Graph;

/**
 * Represents a 2-club solver. An implementation of this Solver will find (if it exists) a size at
 * least k 2-club in the given graph that contains the requiredVertices.
 * 
 * @author sepphar
 * 
 * @param <V>
 * @param <E>
 */
public abstract class Solver<V, E>
{
	public static boolean verbose = false;

	/**
	 * The higher the level the more information the program will provide.
	 */
	public static int verboseLevel = 1;

	private final Graph<V, E> graph;

	private final Set<V> requiredVertices;

	private int twoClubMinSizeParam;

	public Solver(final Graph<V, E> graph, final int k, Collection<V> reqVertices)
	{
		this.graph = graph;
		this.twoClubMinSizeParam = k;
		this.requiredVertices = new HashSet<V>();
		if ( reqVertices != null )
		{
			this.getRequiredVertices().addAll(reqVertices);
		}
	}

	public void addRequiredVertices(Collection<V> c)
	{
		this.requiredVertices.addAll(c);
	}

	public Graph<V, E> getGraph()
	{
		return this.graph;
	}

	public Set<V> getRequiredVertices()
	{
		return this.requiredVertices;
	}

	public int getTwoClubMinSizeParam()
	{
		return this.twoClubMinSizeParam;
	}

	public void setTwoClubMinSizeParam(int size)
	{
		this.twoClubMinSizeParam = size;
	}

	/**
	 * If there is 2-club in the graph of size at least k, then the corresponding vertex set will be
	 * returned by this method and otherwise null.
	 * 
	 * @return
	 * @throws UnsolvableException
	 */
	public abstract Set<V> solve() throws UnsolvableException;
}
