package solvers;

import graph.ConflictCachingGraph;
import graph.GraphAnalysis;
import graph.VertexMaskedSubgraph;

import java.util.Collection;
import java.util.Set;
import java.util.Vector;

import org.jgrapht.Graph;

import solvers.dataReduction.AdvancedRuleProvider;
import solvers.dataReduction.SimpleRuleProvider;
import solvers.turingKernels.TuringKernelProvider;
import solvers.turingKernels.TuringKernelProviderQueue;
import solvers.turingKernels.VertexPriority;
import dataStructures.Store;

public class TuringKernelSolver<V, E> extends Solver<V, E>
{
	public final static boolean turingKernelsVerbose = false;

	public static final int upperBound = 1000;

	public static final double switchToUnderlyingSolver = 0.5;

	public static String getTuringKernelProviderName(int i)
	{
		if ( i == 0 )
			return "MinTwoNeighSize";
		else if ( i == 1 )
			return "MinFESTwoNeigh";
		else
			return "MinDegree";
	}

	private boolean findLargestTwoClub;

	private Set<V> globalMaxTwoClub = null;

	private VertexMaskedSubgraph<V, E> graphT = null;

	public int numberTuringKernels = 0;

	public double time = 0d;

	public final boolean useILPSolver;

	public TuringKernelSolver(Graph<V, E> graph, int mintwoClubSize, boolean findLargestTwoClub, Collection<V> requiredVertices,
			boolean useILPToSolveTuringKernels)
	{
		super(graph, mintwoClubSize, requiredVertices);
		this.findLargestTwoClub = findLargestTwoClub;
		this.useILPSolver = useILPToSolveTuringKernels;
	}

	/**
	 * 
	 * @param graph
	 * @param minTwoClubSize
	 * @param requiredVertices
	 * @param recursivelyRemoveVerticesWhenCreatingTuringKernels
	 */
	public TuringKernelSolver(Graph<V, E> graph, int minTwoClubSize, Collection<V> requiredVertices)
	{
		this(graph, minTwoClubSize, false, requiredVertices, false);
	}

	@Override
	public Graph<V, E> getGraph()
	{
		if ( this.graphT != null )
			return this.graphT;
		return super.getGraph();
	}

	public VertexMaskedSubgraph<V, E> getGraphT()
	{
		return this.graphT;
	}

	private void reduceGraph() throws UnsolvableException
	{

		if ( (this.graphT == null || this.graphT instanceof VertexMaskedSubgraph)
				&& this.getGraph().vertexSet().size() <= TuringKernelSolver.upperBound )
		{
			this.graphT = new ConflictCachingGraph<V, E>(this.getGraph(), this.getGraph().vertexSet());
		}
		else if ( this.graphT == null )
		{
			this.graphT = new VertexMaskedSubgraph<V, E>(this.getGraph(), this.getGraph().vertexSet());
		}

		if ( this.graphT instanceof ConflictCachingGraph )
		{
			DualSearchTreeSolver.applyDataReductionRules(new AdvancedRuleProvider<V, E>(this,
					(ConflictCachingGraph<V, E>) this.graphT));
		}
		else
		{
			DualSearchTreeSolver.applyDataReductionRules(new SimpleRuleProvider<V, E>(this, this.graphT));
		}
	}

	@Override
	public Set<V> solve() throws UnsolvableException
	{
		return this.solve(0);
	}

	/**
	 * Creates a Turing kernel provider and than calls to normal solve method.
	 * 
	 * @throws UnsolvableException
	 */
	public Set<V> solve(int i) throws UnsolvableException
	{
		class MinFESTwoNeigh implements VertexPriority<V>
		{
			@Override
			public int getVertexPriority(V vertex)
			{
				final Set<V> twoNeigh = GraphAnalysis.getNeighborsWithinDistanceS(vertex, TuringKernelSolver.this.getGraphT(), 2);
				int noEdges = 0;
				for (V v : twoNeigh)
				{
					for (V n : TuringKernelSolver.this.getGraphT().getNeighborsOf(v))
						if ( twoNeigh.contains(n) )
							noEdges++;
				}
				return noEdges / 2;
			}
		}

		class MinTwoNeighSize implements VertexPriority<V>
		{
			@Override
			public int getVertexPriority(V vertex)
			{
				return GraphAnalysis.getNeighborsWithinDistanceS(vertex, TuringKernelSolver.this.getGraphT(), 2).size();
			}
		}

		class MinDegreeSize implements VertexPriority<V>
		{
			@Override
			public int getVertexPriority(V vertex)
			{
				return TuringKernelSolver.this.getGraphT().getDegreeOf(vertex);
			}

		}

		final TuringKernelProvider<V, E> provider;
		if ( i == 0 )
		{
			provider = new TuringKernelProviderQueue<V, E>(this, new MinTwoNeighSize());
		}
		else if ( i == 1 )
		{
			provider = new TuringKernelProviderQueue<V, E>(this, new MinFESTwoNeigh());
		}
		else
		{
			provider = new TuringKernelProviderQueue<V, E>(this, new MinDegreeSize());
		}

		return this.solve(provider);
	}

	public Set<V> solve(final TuringKernelProvider<V, E> provider) throws UnsolvableException
	{
		{// check max degree 2-club
			final Set<V> tmp = GraphAnalysis.getMaxDegreeTwoClub(this.getGraph()).getS2();
			this.globalMaxTwoClub = tmp;
			if ( this.getTwoClubMinSizeParam() <= tmp.size() )
			{
				if ( !this.findLargestTwoClub )
					return this.globalMaxTwoClub;
				else
				{
					this.setTwoClubMinSizeParam(this.globalMaxTwoClub.size() + 1);
				}
			}
		}
		if ( Solver.verbose )
			System.out.println("checked max degree 2-club:" + this.globalMaxTwoClub.size());

		// change graph representation and apply data reduction
		try
		{
			this.reduceGraph();
		} catch (final UnsolvableException e1)
		{
			return this.globalMaxTwoClub;
		}

		if ( Solver.verbose )
			System.out.println("initially reduced the graph");

		Store<V, Set<V>> store;
		while ((store = provider.getNextTuringKernel()) != null)
		{
			if ( Solver.verbose )
			{
				System.out.println("#vertices remaining:" + this.graphT.vertexSet().size() + "\t#turingkernels:"
						+ this.numberTuringKernels);
			}

			if ( store.getS2().size() > switchToUnderlyingSolver * this.graphT.vertexSet().size() )
			{
				final Solver<V, E> solv = (this.useILPSolver) ? new ILPSolver<V, E>(this.graphT, this.getTwoClubMinSizeParam(),
						this.getRequiredVertices()) : new DualSearchTreeSolver<V, E>(this.graphT, this.getTwoClubMinSizeParam(),
						this.getRequiredVertices(), this.findLargestTwoClub);
				final Set<V> t = solv.solve();
				return t != null ? t : this.globalMaxTwoClub;
			}

			final Vector<V> req = new Vector<V>(this.getRequiredVertices());
			req.add(store.getS1());

			final long s = System.currentTimeMillis();

			final ConflictCachingGraph<V, E> turingGraph = new ConflictCachingGraph<V, E>(this.graphT, store.getS2());
			final Set<V> twoClub;

			if ( this.useILPSolver )
			{
				twoClub = new ILPSolver<V, E>(turingGraph, this.getTwoClubMinSizeParam(), req).solve();
			}
			else
			{
				twoClub = new DualSearchTreeSolver<V, E>(turingGraph, this.getTwoClubMinSizeParam(), req, this.findLargestTwoClub)
						.solve();
			}

			this.time += ((System.currentTimeMillis() - s) / 1000d);
			this.numberTuringKernels++;

			if ( twoClub != null )
			{
				if ( !this.findLargestTwoClub )
					return twoClub;

				this.setTwoClubMinSizeParam(twoClub.size() + 1);
				this.globalMaxTwoClub = twoClub;
			}

			provider.removeVertex(store.getS1());

			try
			{
				this.reduceGraph();
			} catch (final UnsolvableException e)
			{
				return this.globalMaxTwoClub;
			}

		}
		return this.globalMaxTwoClub;
	}
}