package solvers.dataReduction;

import graph.ConflictCachingGraph;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import solvers.UnsolvableException;

public class HighDegreeRule<V, E> implements RuleInterface<V, E>
{

	class VerticesToBeRemoved extends HashSet<V>
	{
		private final Set<V> req;

		private static final long serialVersionUID = -627480603412377789L;

		public VerticesToBeRemoved(Set<V> requiredVertices)
		{
			super();
			this.req = requiredVertices;
		}

		public void addAllRemovableVertex(Collection<V> conflictSet) throws UnsolvableException
		{
			Collection<V> t = this.req;
			Collection<V> l = conflictSet;
			if ( conflictSet.size() < this.req.size() )
			{
				t = conflictSet;
				l = this.req;
			}

			for (final V v : t)
			{
				if ( l.contains(v) )
					throw new UnsolvableException();
			}

			super.addAll(conflictSet);
		}

		public boolean addRemovableVertex(V e) throws UnsolvableException
		{
			if ( this.req.contains(e) )
				throw new UnsolvableException();

			return super.add(e);
		}
	}

	private AdvancedRuleProvider<V, E> adv;

	/**
	 * Provides a lower bound on the number of vertices that have to be deleted in order to solve
	 * all conflicts within the given vertex set.
	 * 
	 * @param vertexSet
	 * @param conflictMapping
	 * @param minValue
	 * 
	 * @return
	 * @throws UnsolvableException
	 */
	private int getLowerBound(ConflictCachingGraph<V, E> graph, Set<V> vertexSet, Set<V> requiredVertices, int minValue)
			throws UnsolvableException
	{
		final Set<V> vertexCover = new HashSet<V>();
		final VerticesToBeRemoved removeVertices = new VerticesToBeRemoved(requiredVertices);

		{// check all conflicts with required vertices
			final Iterator<V> iterReq = requiredVertices.iterator();

			while (iterReq.hasNext() && vertexCover.size() < minValue)
			{
				final V reqV = iterReq.next();
				final Set<V> conflictsWithreqV = this.adv.getConflictMapping().get(reqV);
				vertexCover.addAll(conflictsWithreqV);
				removeVertices.addAllRemovableVertex(conflictsWithreqV);
			}
		}

		int lw = vertexCover.size();
		if ( lw >= minValue )
			return lw;

		for (final V v1 : vertexSet)
		{
			// degree one vertex rule || twoNeighborhood to small
			if ( graph.getDegreeOf(v1) == 1 || this.adv.getTwoNeighSize(v1) < this.adv.getTwoClubMinSize() )
			{
				removeVertices.addRemovableVertex(v1);
				if ( !vertexCover.contains(v1) )
				{
					vertexCover.add(v1);
					lw++;
				}
			}
			else if ( !vertexCover.contains(v1) && !requiredVertices.contains(v1) )
			{
				final Set<V> conflictSetForV1 = graph.getConflictMapping().get(v1);

				int numberNonVertexCoverNeighbors = 0;
				V nonVertexCoverNeighbor = null;

				{
					int i = 0;
					for (final V v2Con : conflictSetForV1)
					{
						i++;
						if ( !vertexCover.contains(v2Con) )
						{
							numberNonVertexCoverNeighbors++;
							nonVertexCoverNeighbor = nonVertexCoverNeighbor == null
									|| graph.getConflictMapping().get(nonVertexCoverNeighbor).size() > graph.getConflictMapping()
											.get(v2Con).size() ? v2Con : nonVertexCoverNeighbor;

							if ( numberNonVertexCoverNeighbors + lw >= minValue
									|| lw + numberNonVertexCoverNeighbors + (conflictSetForV1.size() - i) < minValue )
							{
								break;
							}
						}
					}
				}

				if ( numberNonVertexCoverNeighbors + lw >= minValue )
				{
					removeVertices.add(v1);
					vertexCover.add(v1);
					lw++;
				}
				else if ( nonVertexCoverNeighbor != null )
				{
					vertexCover.add(v1);
					vertexCover.add(nonVertexCoverNeighbor);
					lw++;
				}
			}
			if ( lw >= minValue )
				return lw;
		}

		if ( removeVertices.size() > 0 )
		{
			for (final V v : removeVertices)
			{
				this.adv.removeVertex(v);
			}
			return Integer.MAX_VALUE;
		}
		return -1;
	}

	@Override
	public boolean reduce(SimpleRuleProvider<V, E> provider) throws UnsolvableException
	{
		if ( provider instanceof AdvancedRuleProvider )
		{
			this.adv = (AdvancedRuleProvider<V, E>) provider;
			final int min = provider.vertexSet().size() - provider.getTwoClubMinSize() + 1;
			final int lowerBound = this.getLowerBound(this.adv.getGraph(), provider.vertexSet(), provider.getRequiredVertices(),
					min);
			if ( lowerBound == Integer.MAX_VALUE )
				return true;
			else if ( lowerBound >= min )
				throw new UnsolvableException();
		}
		return false;
	}
}
